;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;
;;;     			     ANSIL
;;; 		      Advanced Network Simulator In Lisp
;;; 		      ----------------------------------
;;;
;;;				  Written By
;;;			       Peter J. Angeline
;;;			      Gregory M. Saunders
;;;
;;;			   The Ohio State University
;;;		       Laboratory for AI Research (LAIR)
;;;			     Columbus, Ohio, 43210
;;;
;;;			   Copyright (c) 1991, 1992
;;;
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; ---------------------------------------------------------------------- 
;;; File: generic.lisp - Collection of generic functions that support the
;;; other files.
;;;
;;; Author: Gregory M Saunders
;;;
;;; License: This code may be distributed free of charge.  This code may not be
;;; incorporated into any production software of any type or any code which is
;;; sold without the express written consent of the authors.  Modifications and
;;; inclusions for other types of software are permitted, provided that this
;;; header remains unchanged and included in all subsequent versions of the
;;; code.
;;;
;;; Disclaimer: The authors of this code make no claims as to its
;;; validity, and are in no way responsible for any errors contained within.
;;; 
;;; File description: This file contains an implementation of back prop using
;;; linear algebra routines rather than symbolic means.  
;;;
;;; Creation date: March 11, 1992, Version 2.01
;;; ----------------------------------------------------------------------
;;;
;;;
;;;                         Change Log
;;;                         ----------
;;;
;;; ----------------------------------------------------------------------

;;; ---------------------------------------------------------------------------
;;;
;;; Constants
;;;
;;; ---------------------------------------------------------------------------

(defconstant *element-type* 'double-float)
(defconstant one (coerce 1.0 *element-type*))
(defconstant zero (coerce 0.0 *element-type*))

(defun coerce-vector (r &aux (element-type *element-type*))
  (coerce
   (loop for x in (coerce r 'list) collect
	 (coerce x element-type))
   `(simple-array ,element-type)))

(defun coerce-number (x)
  (coerce x *element-type*))




;;; ---------------------------------------------------------------------------
;;;
;;; Generic functions
;;;
;;; ---------------------------------------------------------------------------

;(defun ith (i x)
;  (cond ((> i (1- (length x))) (error "ith overflow"))
;	((listp x)   (nth i x))
;	((vectorp x) (aref x i))
;	(t (error "ith"))))
;
;(defsetf ith (i x) (value)
;  `(if (listp ,x) (setf (nth ,i ,x) ,value)
;     (setf (aref ,x ,i) ,value)))

(defun list-or-vector (r)
  (or (listp r)
      (vectorp r)))

(defun non-array-atom (x)
  (and (atom x)
       (not (arrayp x))))

(defun square (x)
  (* x x))

(defun matrixp (x)
  (and (arrayp x)
       (not (vectorp x))))

(defun extract (vec &key (start 0) (end (1- (length vec))))
  (coerce-vector
   (loop for k from start to end collect
	 (aref vec k))))


;;; ---------------------------------------------------------------------------
;;;
;;; A couple routines from linear common to both C & lisp.
;;;
;;; ---------------------------------------------------------------------------

;;; create matrix or vector
(defun create-array (dims &key (initial-element 0))
  (make-array dims
	      :element-type *element-type*
	      :initial-element (coerce initial-element *element-type*)))

(defun identity-matrix (n &aux (m (create-array (list n n))))
  (loop for i from 0 to (1- n) do
	(setf (aref m i i) one))
  m)

;; returns a vector with the given item consed onto the front of it.
;(defun cons-vector (item vect)
;  (coerce-vector (cons item (coerce vect 'list))))

(defun append.vecs (&rest vecs &aux res (i 0))
  (setf res (create-array (loop for v in vecs sum (length v))))
  (loop for v in vecs do
	(loop for x across v do
	      (setf (aref res i) x)
	      (incf i)))
  res)

(defun depth (r)
  (if (atom r) 0
    (1+ (loop for x in r maximize (depth x)))))

(defun max-depth (trees)
  (loop for x in trees maximize (depth x)))


;;; ---------------------------------------------------------------------------
;;;
;;; Generic show routines
;;;
;;; ---------------------------------------------------------------------------

(defun show-vector (v)
  (format t "~a" (format-vector v)))

(defun format-vector (v &optional (f$ "#(~{~6,2f~})"))
  (format nil f$ (coerce v 'list)))


;;; ---------------------------------------------------------------------------
;;;
;;; Calling functions with default parameters
;;;
;;; ---------------------------------------------------------------------------

(defun sapply (fn &rest parms)
  (apply fn (splice-last-element parms)))

(defun splice-last-element (r)
  (append
   (loop for i from 0 to (- (length r) 2)
	 for x = (nth i r)
	 collect x)
   (car (last r))))

(defun remove-keywords (parms &rest badwords)
  (loop for i from 0 to (1- (length parms)) by 2
	for key = (nth i parms)
	for val = (nth (1+ i) parms)
	when (not (member key badwords))
	append
	(list key val)))
