;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;
;;;     			     ANSIL
;;; 		      Advanced Network Simulator In Lisp
;;; 		      ----------------------------------
;;;
;;;				  Written By
;;;			       Peter J. Angeline
;;;			      Gregory M. Saunders
;;;
;;;			   The Ohio State University
;;;		       Laboratory for AI Research (LAIR)
;;;			     Columbus, Ohio, 43210
;;;
;;;			   Copyright (c) 1991, 1992
;;;
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; ---------------------------------------------------------------------- 
;;; File: linear1.lisp - linear algebra, lisp version.
;;;
;;; Authors: Peter J Angeline
;;;
;;; License: This code may be distributed free of charge.  This code may not be
;;; incorporated into any production software of any type or any code which is
;;; sold without the express written consent of the authors.  Modifications and
;;; inclusions for other types of software are permitted, provided that this
;;; header remains unchanged and included in all subsequent versions of the
;;; code.
;;;
;;; Disclaimer: The authors of this code make no claims as to its
;;; validity, and are in no way responsible for any errors contained within.
;;; 
;;; File description: This file contains implementations of linear algebra
;;; routines which are implemented in C routines and called by Franz Allegro
;;; Common Lisp.  This file is specifc to Franz Allegro Common Lisp and will
;;; not work with other common lisps.  The file Linear.lisp should be used as a
;;; replacement for this file when using a different common lisp. the file with
;;; the C routines, linear-lib.o, should be placed in the same directory.
;;;
;;; Creation date: March 11, 1992, Version 2.01
;;; ----------------------------------------------------------------------
;;; 
;;;                      Change Log
;;;                      ----------
;;;
;;; ----------------------------------------------------------------------


;;; ----------------------------------------------------------------------
;;;
;;; Matrix Routines
;;;
;;; ----------------------------------------------------------------------

;; matrix-multiply algorithm.  If result is non-nil then it must be a matrix of
;; the correct size in which the computation will be done.  If result is
;; non-nil then the function gives no result but simply alters RESULT.

(defun matrix-multiply (result matrix1 matrix2)
  (unless (equal (array-dimension matrix1 1)
		 (array-dimension matrix2 0))
	  (error "MATRIX-MULTIPLY: Matrix sizes don't agree."))
  (when (and result
	     (not (= (array-dimension matrix1 0) (array-dimension result 0)))
	     (not (= (array-dimension matrix2 1) (array-dimension result 1))))
	  (error "MATRIX-MULTIPLY: Result Matrix not correct size."))
  (let ((answer (if result result
		  (let ((type (type-of (aref matrix1 0 0))))
		    (make-array (list (array-dimension matrix1 0)
				      (array-dimension matrix2 1))
				:element-type type))))
	(sum zero) ;gms 01/21
	(width (array-dimension matrix1 1))
	(rows (array-dimension matrix1 0))
	(cols (array-dimension matrix2 1)))
    (loop for i from 0 to (1- rows) do
	  (loop for j from 0 to (1- cols) do
	  (setf sum zero ;gms 01/21
		(aref answer i j)
		(dotimes (k width sum)
		   (incf sum (* (aref matrix1 i k) (aref matrix2 k j)))))))
    (if result nil answer)))

(defun vector-matrix-multiply (res v m &aux
				   (num-rows (array-dimension m 0))
				   (num-cols (array-dimension m 1))
				   (maxv (1- (length v))))
  (unless (equal maxv
		 (1- num-rows))
	  (error "VECTOR-MATRIX-MULTIPLY: Sizes don't agree."))
  (if (and res
	   (not (= num-cols (length res))))
      (error "VECTOR-MATRIX-MULTIPLY: Result vector not correct size."))
  (unless res (setf res (create-array num-cols)))
  (loop for col from 0 to (1- num-cols) do
	(setf (aref res col)
	      (loop for row from 0 to (1- (length v)) sum
		    (* (aref v row)
		       (aref m row col)))))
  res)

(defun matrix-vector-multiply (res m v &aux
				   (num-rows (array-dimension m 0))
				   (num-cols (array-dimension m 1))
				   (maxv (1- (length v))))
  (unless (equal maxv
		 (1- num-cols))
	  (error "VECTOR-MATRIX-MULTIPLY: Sizes don't agree."))
  (if (and res
	   (not (= num-rows (length res))))
      (error "VECTOR-MATRIX-MULTIPLY: Result vector not correct size."))
  (unless res (setf res (create-array num-rows)))
  (loop for row from 0 to (1- num-rows) do
	(setf (aref res row)
	      (loop for col from 0 to (1- (length v)) sum
		    (* (aref v col)
		       (aref m row col)))))
  res)

(defun matrix-apply* (result matrix &rest vals)
  (apply #'matrix-apply (append (list result #'* matrix) vals)))

(defun matrix-apply+ (result matrix &rest vals)
  (apply #'matrix-apply (append (list result #'+ matrix) vals)))

;; apply op to each element of MATRIX.  When val is a matrix, the matrices MUST
;; be the same size.  When val is a value, simple do op of this value to each
;; element in the matrix. Returns the finished product. Op should be any
;; binary op which returns the correct type of value.
(defun matrix-apply (result op matrix &rest vals)
  (let* ((rows (array-dimension matrix 0))
	 (cols (array-dimension matrix 1))
	 (type (type-of (aref matrix 0 0)))
	 (answer (if result result
		   (make-array (array-dimensions matrix)
			       :element-type type))))
  ; when vals is null, then do a single loop, otherwise move through the list
  ; stored in vals.  The (or vals '(t)) returns '(t) when vals is null so the
  ; loops does at least one iteration.
    (loop for val in (or vals '(nil))
	  for mat = matrix then answer do
	  (when (and (listp (type-of val))
		     (not (equal (array-dimensions mat)
				 (array-dimensions val))))
		(error "MATRIX-APPLY: Matrices must be the same size."))
	  (loop for i from 0 to (1- rows) do
		(loop for j from 0 to (1- cols) do
		      (setf (aref answer i j)
			    (coerce 
			     (apply op 
				    (cons (aref mat i j)
					  (cond ((null vals) nil)
						((listp (type-of val))
						 (list (aref val i j)))
						(t (list val)))))
			     type)))))
    (if result nil answer)))

;; copies a matrix and it's contents into a distinct matrix.
(defun copy-matrix (matrix)
  (let ((result (make-array (array-dimensions matrix)
			    :element-type (type-of (aref matrix 0 0)))))
    (dotimes (i (array-dimension matrix 0) result)
       (dotimes (j (array-dimension matrix 1))
	  (setf (aref result i j) (aref matrix i j))))))

;; performs a dot (inner) product on the given vectors. Result is a single
;; value. Vectors must be of the same length.
(defun dot-product (vector1 vector2)
  (unless (= (array-dimension vector1 0) (array-dimension vector2 0))
	  (error "DOT-PRODUCT: Vectors not the same size."))
  (loop for i from 0 to (1- (length vector1)) 
	sum (* (aref vector1 i) (aref vector2 i))))

;; performs an outer product on the vectors.  Result is a (length vector1) x
;; (length vector2) matrix.
(defun vector-product (result vector1 vector2)
  (when (and result (not (equal (array-dimensions result)
				(list (length vector1) (length vector2)))))
	(error "VECTOR-PRODUCT: Result matrix worng size."))
  (let ((answer (if result result
		  (make-array (list (length vector1) (length vector2))
			    :element-type (type-of (aref vector1 0))))))
    (loop for i from 0 to (1- (length vector1)) 
	  for v1 = (aref vector1 i) do
	  (loop for j from 0 to (1- (length vector2)) do
		(setf (aref answer i j) (* v1 (aref vector2 j)))))
    (if result nil answer)))

(defun vector-apply* (result vect &rest vals)
  (apply #'vector-apply (append (list result #'* vect) vals)))

(defun vector-apply+ (result vect &rest vals)
  (apply #'vector-apply (append (list result #'+ vect) vals)))


;; Applies the operator op to the input vector.  If vals is null, then the op
;; is considered to be unary.  Otherwise, the function applies the operator on
;; left to right pairs so that (vector-apply #'+ v1 v2 v3 v4 v5) => 
;; (+ (+ (+ (+ v1 v2) v3) v4) v5).  All vector must be the same length. When a
;; value in vals is a scaler, then that scaler is applied to each element of
;; the vector with op.  The function returns a new vector.
(defun vector-apply (result op vect &rest vals)
  (let* ((len (array-dimension vect 0))
	 (type (type-of (aref vect 0)))
	 (answer (if result result
		   (make-array (array-dimensions vect)
			       :element-type type))))
  ; when vals is null, then do a single loop, otherwise move through the list
  ; stored in vals.  The (or vals '(t)) returns '(t) when vals is null so the
  ; loops does at least one iteration.
    (loop for val in (or vals '(nil))
	  for ans = vect then answer do
	  (when (and (listp (type-of val))
		     (not (equal (array-dimensions ans)
				 (array-dimensions val))))
		(error "VECTOR-APPLY: Vectors must be the same size."))
	  (dotimes (i len)
		   (setf (aref answer i)
			 (coerce 
			  (apply op 
				 (cons (aref ans i)
				       (cond ((null vals) nil)
					     ((listp (type-of val))
					      (list (aref val i)))
					     (t (list val)))))
			  type))))
    (if result nil answer)))

;; computes the euclidean distance between two vectors.
(defun vector-distance (vector1 vector2)
  (unless (= (length vector1) (length vector2))
	  (error "VECTOR-DISTANCE: Vectors must be of same length."))
  (sqrt (apply #'+ (mapcar #'(lambda (v1 v2) (* (- v1 v2) (- v1 v2)))
			   (coerce vector1 'list)
			   (coerce vector2 'list)))))

;; copies a vector
(defun copy-vector (vec &aux (res (create-array (length vec))))
  (move-vector vec res)
  res)

(defun move-vector (v1 v2 &key fixed-nodes)
  (loop for i from 0 to (1- (length v1))
	do
	(if (or (not fixed-nodes)
		(zerop (aref fixed-nodes i)))
	    (setf (aref v2 i) (aref v1 i)))))

(defun offset-move.vv (from to offset)
  (if (> (+ offset (length from)) (length to)) (error "offset-move.vv"))
  (loop for i from 0 to (1- (length from))
	for x = (aref from i)
	do
	(setf (aref to (+ i offset)) x)))

;(defun offset-move.vv (from to offset)
;  (move.vv from 
;       (make-array (- (length to) offset)
;		   :element-type *element-type*
;		   :displaced-to to
;		   :displaced-index-offset offset)))
      
(defun move-matrix (m1 m2 &aux (d (array-dimensions m1)))
  (loop for i from 0 to (1- (car d)) do
	(loop for j from 0 to (1- (cadr d)) do
	      (setf (aref m2 i j) (aref m1 i j)))))


;;; ---------------------------------------------------------------------------
;;;
;;; Switching between arrays & vectors
;;;
;;; ---------------------------------------------------------------------------

;; returns a row array which is equivalent to the vector. A copy is not made so
;; destructive operations on the result will change the original vector.
(defun row-array (vect)
  (make-array (cons 1 (array-dimensions vect)) :displaced-to vect :element-type (type-of (aref vect 0))))

;; returns a column array which is equivalent to the given vector.  A copy is
;; not made so destructive operations on the result will change the original
;; vector. 
(defun column-array (vect)
  (make-array (append (array-dimensions vect) '(1))
	      :displaced-to vect
	      :element-type (type-of (aref vect 0))))


;;; ---------------------------------------------------------------------------
;;;
;;; Squashing
;;;
;;; ---------------------------------------------------------------------------

(defun sigmoid (x)
  (/ one (+ one (exp (- x)))))

(defun lisp-squash (x &optional (fn one))
  (cond ((= fn 0d0) x)
	((= fn 1d0) (sigmoid x))
	((= fn 2d0) (tanh x))		;(tanh (/ x 1d0))
	(t (error "squash"))))

(defun lisp-dsquash (x &optional (fn one))
  (cond ((= fn 0d0) zero)      
	((= fn 1d0) (* x (- one x)))
	((= fn 2d0) (- one (* x x)))   ;(/ (- one (* x x)) 1d0)
	(t (error "dsquash"))))

(defun squash-vector (res x squashvec)
  (unless res (setf res (create-array (length x))))
  (loop for i from 0 to (1- (length x)) do
	(setf (aref res i)
	      (lisp-squash (aref x i) (aref squashvec i))))
  res)

(defun dsquash-vector (res vec squashvec)
  (unless res (setf res (create-array (length vec))))
  (loop for i from 0 to (1- (length vec)) do
	(setf (aref res i)
	      (lisp-dsquash (aref vec i) (aref squashvec i))))
  res)


;;; ---------------------------------------------------------------------------
;;;
;;; Fuzzy equivalence
;;;
;;; ---------------------------------------------------------------------------

(defun same-vector (v1 v2 &key (threshold zero))
  (loop for x across v1
	for y across v2
	always
	(same-number x y :threshold threshold)))

(defun same-matrix (m1 m2 &key (threshold zero) &aux (d (array-dimensions m1)))
  (loop for i from 0 to (1- (car d)) always
	(loop for j from 0 to (1- (cadr d)) always
	      (same-number (aref m1 i j) (aref m2 i j) :threshold threshold))))

(defun same-number (x y &key (threshold zero))
  (<= (abs (- x y)) threshold))
