/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
/*									  */
/*     			            ANSIL				  */
/* 		      Advanced Network Simulator In Lisp		  */
/* 		      ----------------------------------		  */
/*									  */
/*				  Written By				  */
/*			       Peter J. Angeline			  */
/*			      Gregory M. Saunders			  */
/*									  */
/*			   The Ohio State University			  */
/*		       Laboratory for AI Research (LAIR)		  */
/*			     Columbus, Ohio, 43210			  */
/*									  */
/*			      Copyright (c) 1991			  */
/*									  */
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* ----------------------------------------------------------------------------	*/
/*										*/
/* License: This code may be distributed free of charge.  This code may not be	*/
/* incorporated into any production software of any type or any code which is	*/
/* sold without the express written consent of the authors.  Modifications and	*/
/* inclusions for other types of software are permitted, provided that this	*/
/* header remains unchanged and included in all subsequent versions of the	*/
/* code.									*/
/*										*/
/* Disclaimer: The authors of this code make no claims as to its		*/
/* validity, and are in no way responsible for any errors contained within.	*/
/*										*/
/* File description: This file contains implementations of linear algebra	*/
/* routines used by lisp.  It is in C because it's a bunch faster this way.	*/
/*										*/
/* Creation date: March 17, 1992 Version: 2.01					*/
/* ---------------------------------------------------------------------------- */

/*****************************************************************************/
/* Compile with: cc linear2.c -c -lm -o linear2.o                            */
/*									     */
/* set of functions to run the linear code in the linear.lisp.  This should  */
/* be  a library which is connected to the lisp via the foriegn function     */
/* interface. This should be called in the linear package or the user        */
/* should load the linear package from lisp.                                 */
/*									     */
/* Here's the naming convention:					     */
/*									     */
/*   fn_ab - where fn is the function name, and a & b are one letter	     */
/*     abbreviations of the argument types, ranging over		     */
/*									     */
/*       s = scalar							     */
/*	 v = vector							     */
/*	 m = 2d-matrix							     */
/*	 x = vector or matrix (any dimensionality)			     */ 
/*									     */
/*    So, for example, we have the following:				     */
/*									     */
/*       add_xs      ===> [vector or matrix] * scalar			     */
/*	 multiply_vv ===> vector * vector (pairwise multiplication)	     */
/*       multiply_mm ===> matrix * matrix (pairwise multiplication)	     */
/*									     */
/*   Higher-level multiplications are specified by "dot":		     */
/*									     */
/*       dot_vv ===> dot-product(vector,vector)				     */
/*	 dot_mm ===> matrix-product(matrix,matrix)			     */
/*									     */
/*   Unfortunately, there are two higher-level multiplications for two	     */
/*     vectors:  dot-product & outer-product.  Thus, the outer product is    */
/*									     */
/*       outer_vv ===> outer-product(vector,vector)			     */
/*****************************************************************************/



/***********/
/* Include */
/***********/

#include <stdio.h>
#include <math.h>


/*************/
/* Constants */
/*************/

#define zero ((double) 0.0)
#define one  ((double) 1.0)
#define two  ((double) 2.0)
#define four ((double) 4.0)


/*********************/
/* Generic functions */
/*********************/

/* double square (x) */
/* double x; */
/* {  */
/*   return x*x; */
/* } */


/***************************/
/* Multiplication routines */
/***************************/

/*   void multiply_vs (result, vector, val, len)   */
/*   double *result, *vector, val;   */
/*   int len;   */
/*   {   */
/*     int i;   */
/*     for (i=0;i<len;i++) result[i]=vector[i]*val;   */
/*   }   */

/*   void multiply_vv (result, vector1, vector2, len)   */
/*   double *result, *vector1, *vector2;   */
/*   int len;   */
/*   {   */
/*     int i;   */
/*     for (i=0;i<len;i++) result[i]=vector1[i]*vector2[i];   */
/*   }   */

void multiply_xs (result, x, val, len)
double *result, *x, val;
int len;
{
  int i;
  for (i=0;i<len;i++)
    result[i]=x[i]*val;
}

void multiply_xx (result, x1, x2, len)
double *result, *x1, *x2;
int len;
{
  int i;
  for (i=0;i<len;i++)
    result[i]=x1[i]*x2[i];
}


/*********************/
/* Addition routines */
/*********************/

/*   void add_vs (result, vector, val, len)   */
/*   double *result, *vector, val;   */
/*   int len;   */
/*   {   */
/*     int i;   */
/*     for (i=0;i<len;i++) result[i]=vector[i]+val;   */
/*   }   */
/*      */
/*   void add_vv (result, vector1, vector2, len)   */
/*   double *result, *vector1, *vector2;   */
/*   int len;   */
/*   {   */
/*     int i;   */
/*     for (i=0;i<len;i++) result[i]=vector1[i]+vector2[i];   */
/*   }   */

void add_xs (result, x, val, len)
double *result, *x, val;
int len;
{
  int i;
  for (i=0;i<len;i++)
    result[i]=x[i]+val;
}

void add_xx (result, x1, x2, len)
double *result, *x1, *x2;
int len;
{
  int i;
  for (i=0;i<len;i++)
    result[i]=x1[i]+x2[i];
}


/**********************************/
/* Dot products & vector products */
/**********************************/

double dot_vv (vector1, vector2, len)
double *vector1,*vector2;
int len;
{
  double result;
  int i;
  result=0.0;
  for (i=0;i<len;i++)
    result+= vector1[i] * vector2[i];
  return result;
}

void outer_vv (result, vector1, vector2, len1, len2)
double *result,*vector1,*vector2;
int len1,len2;
{
  int i,j;
  for (i=0;i<len1;i++)
    for (j=0;j<len2;j++)
      result[i*len2+j] = vector1[i] * vector2[j];
}

void dot_mm (result, matrix1, matrix2, rows, width, cols)
double *result, *matrix1, *matrix2;
int rows, cols, width;
{
  int i,j,k;
  /*  printf ("%d %d %d %d %d %d\n", result, matrix1, matrix2, rows, width, cols); */
  for (i=0;i<rows;i++)
    for (j=0;j<cols;j++)
      {
	result[i*cols+j]= zero;
	for (k=0; k<width; k++)
	  result[i*cols+j] += matrix1[i*width+k] * matrix2[k*cols+j];
      }
}


void dot_vm (result, vec, matrix, m, n)  /* matrix is m x n */
double *result, *matrix, *vec;
int m, n;
{
  int j,k;
    for (j=0;j<n;j++)
      {
	result[j]= zero;
	for (k=0; k<m; k++)
	  result[j] += vec[k] * matrix[k * n + j];
      }
}

void dot_mv (result, matrix, vec, m, n)  /* matrix is m x n */
double *result, *matrix, *vec;
int m, n;
{
  int i,k;
    for (i=0;i<m;i++)
      {
	result[i]= zero;
	for (k=0; k<n; k++)
	  result[i] += vec[k] * matrix[i * n + k];
      }
}


/***********************/
/* Squashing routines  */
/***********************/

double tanh();

double sigmoid (x)
double x;
{
  double exp();
  return (one / (one + exp(-x)));
}

double squash (x, fn)
double x, fn;
{
  switch (fn)
    {
    case 0: { return x; }
    case 1: { return sigmoid(x); }
    case 2: { return tanh(x); }
    default: { return x; }
    }
}

double dsquash (x, fn)
double x, fn;
{
  switch (fn)
    {
    case 0: { return zero; }
    case 1: { return x * (one - x); }
    case 2: { return one - (x*x); }
    default: { return zero; }
    }
}

void squash_x (res, vector, fn, len)
double *res, *vector, *fn;
int len;
{
  double squash();
  int i;
  for (i=0;i<len;i++)
    res[i]=squash(vector[i],fn[i]);
}

void dsquash_x (res, vector, fn, len)
double *res, *vector, *fn;
int len;
{
  double dsquash();
  int i;
  for (i=0;i<len;i++)
    res[i]=dsquash(vector[i],fn[i]);
}


/************/
/* Equality */
/************/

/*    printf ("%d %d %d %d\n", x1[i],x2[i], abs(x1[i]-x2[i]), threshold);   */

double equal_xx (x1, x2, len, threshold)
double *x1,*x2, threshold;
int len;
{
  double result;
  double fabs();
  int i;
  for (i=0;i<len;i++)
    if ( fabs( x1[i] - x2[i] ) > threshold )  return (one); 
  return (zero);
}

/*****************/
/* Misc routines */
/*****************/

double dist_vv (vector1, vector2, len)    /* euclidean */
double *vector1,*vector2;
int len;
{
  double result;
  double sqrt();
  int i;
  result=0.0;
  for (i=0;i<len;i++)
    result+= (vector1[i] - vector2[i]) * (vector1[i] - vector2[i]);
  return ( sqrt(result));
}

void copy_x (result, x, len)
double *x,*result;
int len;
{
  int i;
  for (i=0;i<len;i++)
    result[i] = x[i];
}

void move_xx (vector1, vector2, fixed_nodes, len)
double *vector1, *vector2, *fixed_nodes;
int len;
{     
  int i;
  for (i=0;i<len;i++)
    if (fixed_nodes[i] == 0) 
  {    vector2[i] = vector1[i]; }
}

void offset_move_xx (vector1, vector2, offset, len)
double *vector1, *vector2;
int offset, len;
{     
  int i;
  for (i=0;i<len;i++)
    vector2[i+offset] = vector1[i];
}



/***********************************************************************/
/*                                                                     */
/*								       */
/*	NO LONGER NEEDED???					       */
/*								       */
/*								       */
/***********************************************************************/

/* gets the desired row from the matrix and places it in the result  vector */

/* void getrow (result, matrix, rownum, rows, cols)   */
/* double *result,*matrix;   */
/* int rownum, rows, cols;   */
/* {   */
/*   int i;   */
/*   for (i=0;i<cols;i++)   */
/*       result[i]=matrix[rownum*cols+i];   */
/* }   */

/* places desired column in result */

/* void getcolumn (result, matrix, colnum, rows, cols)   */
/* double *result,*matrix;   */
/* int colnum,rows,cols;   */
/* {   */
/*   int i;   */
/*   for (i=0;i<rows;i++)   */
/*     result[i]=matrix[i*cols+colnum];   */
/* }   */

/* changes the rownumth row to be what is in the vector */  

/* void set_row (matrix, rownum, vector, rows, cols)   */
/* double *matrix,*vector;   */
/* int rownum,rows,cols;   */
/* {   */
/*   int i;   */
/*   for (i=0;i<cols;i++)   */
/*     matrix[rownum*cols+i] = vector[i];   */
/* }   */

/* changes the colnumth column of the matrix to be what is stored in the vector */  

/* void set_column (matrix, colnum, vector, rows, cols)   */
/* double *matrix,*vector;   */
/* int colnum, rows,cols;   */
/* {   */
/*   int i;   */
/*   for (i=0;i<rows;i++)   */
/*     matrix[i*cols+colnum] = vector[i];   */
/* }   */

/* compute the midpoint between two vectors */

/* void vector_midpoint (result, vector1, vector2, len)   */
/* double *result, *vector1, *vector2;   */
/* int len;   */
/* {   */
/*   int i;   */
/*   for (i=0; i<len;i++)   */
/*     result[i] = vector1[i] + ((vector2[i] - vector1[i]) / two);   */
/* }   */
