;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;
;;;     			     ANSIL
;;; 		      Advanced Network Simulator In Lisp
;;; 		      ----------------------------------
;;;
;;;				  Written By
;;;			       Peter J. Angeline
;;;			      Gregory M. Saunders
;;;
;;;			   The Ohio State University
;;;		       Laboratory for AI Research (LAIR)
;;;			     Columbus, Ohio, 43210
;;;
;;;			   Copyright (c) 1991, 1992
;;;
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; ---------------------------------------------------------------------- 
;;; File: linear2.lisp - Lisp interface for the C version of linear algebra
;;; functions. 
;;;
;;; Authors: Peter J Angeline, Gregory M Saunders
;;;
;;; License: This code may be distributed free of charge.  This code may not be
;;; incorporated into any production software of any type or any code which is
;;; sold without the express written consent of the authors.  Modifications and
;;; inclusions for other types of software are permitted, provided that this
;;; header remains unchanged and included in all subsequent versions of the
;;; code.
;;;
;;; Disclaimer: The authors of this code make no claims as to its
;;; validity, and are in no way responsible for any errors contained within.
;;; 
;;; File description: This file contains implementations of linear algebra
;;; routines which are implemented in C routines and called by Franz Allegro
;;; Common Lisp.  This file is specifc to Franz Allegro Common Lisp and will
;;; not work with other common lisps.  The file linear1.lisp should be used as
;;; a replacement for this file when using a different common lisp. the file
;;; with the C routines, linear2.o, should be placed in the same directory.
;;;
;;; Naming convention:
;;;     This file is simply the lisp counterpart to the foreign functions of
;;;     linear2.c, so the naming convention is very similar.  C functions are
;;;     designated in lisp by an "&", and the words like "add" in C are
;;;     replaced by a "+".  So, for example, we have the following: 
;;;									     
;;;        &+xs    ===> [vector or matrix] + scalar			     
;;;	   &*vv    ===> vector * vector (pairwise multiplication)	     
;;;        &*mm    ===> matrix * matrix (pairwise multiplication) 
;;;        &.vv    ===> dot-product(vector,vector)				     
;;;	   &.mm    ===> matrix-product(matrix,matrix)			     
;;;        &outer  ===> outer-product(vector,vector)			     
;;;        &copy.x ===> copy [vector or matrix]
;;;
;;;     Note that in some cases (like "&outer") the "vv" is dropped because
;;;     outer products apply only to two vectors -- we skipped the generalized
;;;     outer product in this version.
;;;
;;; Creation date: March 11, 1992, Version 2.01
;;; ----------------------------------------------------------------------
;;;
;;;
;;;                         Change Log
;;;                         ----------
;;;
;;; ----------------------------------------------------------------------


;;; ----------------------------------------------------------------------
;;;
;;; Foreign Function Interface
;;;
;;; ----------------------------------------------------------------------

(ff:reset-entry-point-table)	; erase any entry points so don't double
				; load c code, which causes an error.

(eval-when (load)
	   (load (concatenate 'string *ansil-path* *c-file* ".o") :system-libraries '("m")))

(ff:defforeign-list
 `(

   ;; -----------------------------------------------------------
   ;; Multiplication routines
   ;; -----------------------------------------------------------

   (&*xs
    :arguments ((array double-float) (array double-float) double-float integer)
    :entry-point ,(convert-to-lang "multiply_xs")
    :return-type :void)

   (&*xx
    :arguments ((array double-float) (array double-float) (array double-float) integer)
    :entry-point ,(convert-to-lang "multiply_xx")
    :return-type :void)

   ;; -----------------------------------------------------------
   ;; Addition routines
   ;; -----------------------------------------------------------

   (&+xs
    :arguments ((array double-float) (array double-float) double-float integer)
    :entry-point ,(convert-to-lang "add_xs")
    :return-type :void)

   (&+xx
    :arguments ((array double-float) (array double-float) (array double-float) integer)
    :entry-point ,(convert-to-lang "add_xx")
    :return-type :void)

   ;; -----------------------------------------------------------
   ;; Dot products & vector products
   ;; -----------------------------------------------------------

   (&.vv 
     :arguments ((array double-float) (array double-float) integer)
     :entry-point ,(convert-to-lang "dot_vv")
     :return-type :double-float)

   (&outer
    :arguments ((array double-float) (array double-float) (array double-float) integer integer)
    :entry-point ,(convert-to-lang "outer_vv")
    :return-type :void)

   (&.mm 
     :arguments ((array double-float) (array double-float) (array double-float) integer integer integer)
     :entry-point ,(convert-to-lang "dot_mm")
     :return-type :void)

   (&.vm 
     :arguments ((array double-float) (array double-float) (array double-float) integer integer)
     :entry-point ,(convert-to-lang "dot_vm")
     :return-type :void)

   (&.mv
     :arguments ((array double-float) (array double-float) (array double-float) integer integer)
     :entry-point ,(convert-to-lang "dot_mv")
     :return-type :void)

   ;; -----------------------------------------------------------
   ;; Squashing
   ;; -----------------------------------------------------------

   (&squash 
    :arguments (double-float double-float)
    :entry-point ,(convert-to-lang "squash")
    :return-type :double-float)

   (&dsquash 
    :arguments (double-float double-float)
    :entry-point ,(convert-to-lang "dsquash")
    :return-type :double-float)
   
   (&squash.x
    :arguments ((array double-float) (array double-float) (array double-float) integer) 
    :entry-point ,(convert-to-lang "squash_x")
    :return-type :void)

   (&dsquash.x 
    :arguments ((array double-float) (array double-float) (array double-float) integer)
    :entry-point ,(convert-to-lang "dsquash_x")
    :return-type :void)

   ;; -----------------------------------------------------------
   ;; Equality
   ;; -----------------------------------------------------------

   (&=xx
    :arguments ((array double-float) (array double-float) integer double-float)
    :entry-point ,(convert-to-lang "equal_xx")
    :return-type :double-float)

   ;; -----------------------------------------------------------
   ;; Misc routines
   ;; -----------------------------------------------------------

   (&dist.vv
    :arguments ((array double-float) (array double-float) integer)
    :entry-point ,(convert-to-lang "dist_vv")
    :return-type :double-float)

   (&copy.x
    :arguments ((array double-float) (array double-float) integer)
    :entry-point ,(convert-to-lang "copy_x")
    :return-type :void)

   (&move.xx
    :arguments ((array double-float) (array double-float) (array double-float) integer)
    :entry-point ,(convert-to-lang "move_xx")
    :return-type :void)

   (&offset-move.xx
    :arguments ((array double-float) (array double-float) integer integer)
    :entry-point ,(convert-to-lang "offset_move_xx")
    :return-type :void)))
