;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;
;;;     			     ANSIL
;;; 		      Advanced Network Simulator In Lisp
;;; 		      ----------------------------------
;;;
;;;				  Written By
;;;			       Peter J. Angeline
;;;			      Gregory M. Saunders
;;;
;;;			   The Ohio State University
;;;		       Laboratory for AI Research (LAIR)
;;;			     Columbus, Ohio, 43210
;;;
;;;			   Copyright (c) 1991, 1992
;;;
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; ---------------------------------------------------------------------- 
;;; File: math.lisp - Several mathematica-like lisp routines to make life
;;; simpler.
;;;
;;; Author: Gregory M Saunders
;;;
;;; License: This code may be distributed free of charge.  This code may not be
;;; incorporated into any production software of any type or any code which is
;;; sold without the express written consent of the authors.  Modifications and
;;; inclusions for other types of software are permitted, provided that this
;;; header remains unchanged and included in all subsequent versions of the
;;; code.
;;;
;;; Disclaimer: The authors of this code make no claims as to its
;;; validity, and are in no way responsible for any errors contained within.
;;; 
;;; File description:  The linear routines come in two flavors:  C & lisp.
;;;     This file is a set of macros that serve as an interface to either
;;;     library. Thus the user should never have to call any function in
;;;     linear1.lisp or linear2.lisp -- instead he should just use the
;;;     appropriate macro in this file.
;;;
;;; Naming convention:
;;;
;;;        +xs    ===> [vector or matrix] + scalar			     
;;;	   *vv    ===> vector * vector (pairwise multiplication)	     
;;;        *mm    ===> matrix * matrix (pairwise multiplication) 
;;;        .vv    ===> dot-product(vector,vector)				     
;;;	   .mm    ===> matrix-product(matrix,matrix)			     
;;;        outer  ===> outer-product(vector,vector)			     
;;;        copy.x ===> copy [vector or matrix]
;;;  
;;;     Finally in this file we have two functions (NOT MACROS) that can be
;;;     used to do any type of multiplication/addition/dot.  It is provided for
;;;     the user's convenience when working at the lisp top level.  They should
;;;     be avoided in actual code, since the macros won't be expanded, and this
;;;     will slow things down at run time.
;;;
;;;        (%* a b) ===> multiplication
;;;        (%+ a b) ===> addition
;;;        (%. a b) ===> a dot function...  one of [.mm .vv .mv .vm]
;;;
;;; ----------------------------------------------------------------------
;;; Creation date: March 11, 1992, Version 2.01
;;; ----------------------------------------------------------------------
;;;
;;;
;;;                         Change Log
;;;                         ----------
;;;
;;; ----------------------------------------------------------------------


;;; ---------------------------------------------------------------------------
;;;
;;; Multiplication routines
;;;
;;; ---------------------------------------------------------------------------

(defmacro *vs (v s &key res)
  (if *c* (macroexpand `(*xs ,v ,s :res ,res))
    `(vector-apply* ,res ,v ,s)))

(defmacro *vv (v1 v2 &key res)
  (if *c* (macroexpand `(*xx ,v1 ,v2 :res ,res))
    `(vector-apply* ,res ,v1 ,v2)))

(defmacro *ms (m s &key res)
  (if *c* (macroexpand `(*xs ,m ,s :res ,res))
    `(matrix-apply* ,res ,m ,s)))

(defmacro *mm (m1 m2 &key res)
  (if *c* (macroexpand `(*xx ,m1 ,m2 :res ,res))
    `(matrix-apply* ,res ,m1 ,m2)))

;; The following 2 macros support multiplication by multi-dimensional arrays.
;; This was easy since C just gets a pointer to the structure, which is the
;; same for vectors, 2d-arrays, 3d-arrays, etc.  Lisp versions are difficult
;; because they'd have to have n loop statements for n-array.  Hence, the lisp
;; equivalent does not exist.

(defmacro *xs (m s &key res)
  (if *c* 
      `(let* ((d (array-dimensions ,m))
	      (ans (or ,res (create-array d))))
	 (if (and ,res (not (equal (array-dimensions ,res) d)))
	     (error "*xs -- bad result"))
	 (&*xs ans ,m ,s (apply #'* d))
	 ans)
    (error "undefined macro")))

(defmacro *xx (m1 m2 &key res)
  (if *c*
      `(let* ((d1 (array-dimensions ,m1))
	      (ans (or ,res (create-array d1))))
	 (if (or (not (equal d1 (array-dimensions ,m2)))
		 (and ,res (not (equal d1 (array-dimensions ,res)))))
	     (error "*xx -- bad length"))
	 (&*xx ans ,m1 ,m2 (apply #'* d1))
	 ans)
    (error "undefined macro")))

;;; ---------------------------------------------------------------------------
;;;
;;; Addition routines
;;;
;;; ---------------------------------------------------------------------------

(defmacro +vs (v s &key res)
  (if *c* (macroexpand `(+xs ,v ,s :res ,res))
    `(vector-apply+ ,res ,v ,s)))

(defmacro +vv (v1 v2 &key res)
  (if *c* (macroexpand `(+xx ,v1 ,v2 :res ,res))
    `(vector-apply+ ,res ,v1 ,v2)))

(defmacro +ms (m s &key res)
  (if *c* (macroexpand `(+xs ,m ,s :res ,res))
    `(matrix-apply+ ,res ,m ,s)))

(defmacro +mm (m1 m2 &key res)
  (if *c* (macroexpand `(+xx ,m1 ,m2 :res ,res))
    `(matrix-apply+ ,res ,m1 ,m2)))

(defmacro -mm (m1 m2 &key res)
  (if *c* (macroexpand `(-xx ,m1 ,m2 :res ,res))
    (error "-mm not defined for lisp mode")))


;; The following 2 macros support addition with multi-dimensional arrays.
;; This was easy since C just gets a pointer to the structure, which is the
;; same for vectors, 2d-arrays, 3d-arrays, etc.  Lisp versions are difficult
;; because they'd have to have n loop statements for n-array.  Hence, the lisp
;; equivalent does not exist.

(defmacro +xs (m s &key res)
  (if *c*
      `(let* ((d (array-dimensions ,m))
	      (ans (or ,res (create-array d))))
	 (if (and ,res (not (equal (array-dimensions ,res) d)))
	     (error "*xs -- bad result"))
	 (&+xs ans ,m ,s (apply #'* d))
	 ans)
    (error "undefined macro")))

(defmacro +xx (m1 m2 &key res)
  (if *c*
      `(let* ((d1 (array-dimensions ,m1))
	      (ans (or ,res (create-array d1))))
	 (if (or (not (equal d1 (array-dimensions ,m2)))
		 (and ,res (not (equal d1 (array-dimensions ,res)))))
	     (error "*xx -- bad length"))
	 (&+xx ans ,m1 ,m2 (apply #'* d1))
	 ans)
    (error "undefined macro")))

;;; This isn't so good, since *xs doesn't get expanded ???

(defmacro -xx (x1 x2 &key res)
  (if *c* 
      (macroexpand `(+xx ,x1 
			 ,(macroexpand `(*xs ,x2 -1d0)) :res ,res))
    (error "-xx not defined for lisp mode")))


;;; ---------------------------------------------------------------------------
;;;
;;; Dot products & vector products
;;;
;;; ---------------------------------------------------------------------------

(defmacro .vv (v1 v2)
  (if *c* 
      `(let* ((n (length ,v1)))
	 (unless (= n (length ,v2)) (error ".vv -- bad length"))
	 (&.vv ,v1 ,v2 n))
    `(dot-product ,v1 ,v2)))

(defmacro outer (v1 v2 &key res)
  (if *c*
      `(let* ((n1 (length ,v1))
	      (n2 (length ,v2))
	      (ans (or ,res (create-array (list n1 n2)))))
	 (if (and ,res
		  (or (not (= (length (array-dimensions ,res)) 2))
		      (not (= n1 (array-dimension ,res 0)))
		      (not (= n2 (array-dimension ,res 1)))))
	     (error "outer -- bad length"))
	 (&outer ans ,v1 ,v2 n1 n2)
	 ans)
    `(matrix-multiply ,res (column-array ,v1) (row-array ,v2))))

(defmacro .mm (m1 m2 &key res)
  (if *c*
      `(let* ((mdim (array-dimension ,m1 0))
	      (rdim (array-dimension ,m1 1))
	      (ndim (array-dimension ,m2 1))
	      (ans (or ,res (create-array (list mdim ndim)))))
	 (if (or (not (= rdim (array-dimension ,m2 0)))
		 (and ,res
		      (or (not (= (length (array-dimensions ,res)) 2))
			  (not (= mdim (array-dimension ,res 0)))
			  (not (= ndim (array-dimension ,res 1))))))
	     (error ".mm -- bad length"))
	 (&.mm ans ,m1 ,m2 mdim rdim ndim)
	 ans)
    `(matrix-multiply ,res ,m1 ,m2)))

(defmacro .vm (v m &key res)
  (if *c*
      `(let* ((m! (array-dimension ,m 0))         ; matricies are always M x N!
	      (n! (array-dimension ,m 1))
	      (ans (or ,res (create-array n!))))
	 (if (or (not (= m! (length ,v)))
		 (and ,res (not (= (length ,res) n!))))
	     (error ".vm -- bad length"))
	 (&.vm ans ,v ,m m! n!)
	 ans)
    `(vector-matrix-multiply ,res ,v ,m)))

(defmacro .mv (m v &key res)
  (if *c*
      `(let* ((m! (array-dimension ,m 0))         ; matricies are always M x N!
	      (n! (array-dimension ,m 1))
	      (ans (or ,res (create-array m!))))
	 (if (or (not (= n! (length ,v)))
		 (and ,res (not (= (length ,res) m!))))
	     (error ".mv -- bad length"))
	 (&.mv ans ,m ,v m! n!)
	 ans)
    `(matrix-vector-multiply ,res ,m ,v)))


;;; ---------------------------------------------------------------------------
;;;
;;; Copying & moving
;;;
;;; ---------------------------------------------------------------------------

(defmacro copy.v (v &key res)
  (if *c* (macroexpand `(copy.x ,v :res ,res))
    `(progn
       (if ,res (error "copy.v doesn't support :res in lisp-mode")
	 (copy-vector ,v)))))

(defmacro copy.m (m &key res)
  (if *c* (macroexpand `(copy.x ,m :res ,res))
    `(progn
       (if ,res (error "copy.m doesn't support :res in lisp-mode")
	 (copy-matrix ,m)))))

(defmacro copy.x (x &key res)
  (if *c*
      `(let* ((dims (array-dimensions ,x))
	      (ans (or ,res (create-array dims))))
	 (if (and ,res
		  (not (equal dims (array-dimensions ,res))))
	     (error "copy.x -- bad length"))
	 (&copy.x ans ,x (apply #'* dims))
	 ans)
    (error "copy.x not defined in lisp mode")))

(defmacro move.vv (from to &key fixed-nodes offset)
  (cond ((and offset fixed-nodes) (error "move.vv"))
	(offset 
	 (if *c*
	     `(let ((n (length ,from)))
		(if (or (> (+ ,offset n) (length ,to))
			(< ,offset 0))
		    (error "move.vv -- bad length"))
		(&offset-move.xx ,from ,to ,offset n))
	   `(offset-move.vv ,from ,to ,offset)))
	(*c* `(let ((n (length ,from)))
		(if (or (> n (length ,to))
			(and ,fixed-nodes 
			     (not (= (length ,fixed-nodes) (length ,to)))))
		    (error "move.vv -- bad length"))
		(if ,fixed-nodes
		    (&move.xx ,from ,to ,fixed-nodes n)
		  (&copy.x ,to ,from n))))
	(t `(move-vector ,from ,to :fixed-nodes ,fixed-nodes))))

;(defmacro move.vv (from to &key fixed-nodes)
;  (if *c* 
;      `(let ((n (length ,from)))
;	 (if (or (> n (length ,to))
;		 (and ,fixed-nodes 
;		      (not (= (length ,fixed-nodes) (length ,to)))))
;	     (error "move.vv -- bad length"))
;	 (if ,fixed-nodes
;	     (&move.xx ,from ,to ,fixed-nodes n)
;	   (&copy.x ,to ,from n)))
;    `(move-vector ,from ,to :fixed-nodes ,fixed-nodes)))

;(defmacro move.xx (from to &key fixed-nodes)
;  (if *c*
;      `(let ((d (array-dimensions ,from)))
;	 (unless (equal d (array-dimensions ,to)) (error "move.xx -- bad length"))
;	 (if ,fixed-nodes
;	     (progn
;	       (unless (equal d (array-dimensions ,fixed-nodes))
;		       (error "move.xx -- bad length"))
;	       (&move.xx ,from ,to ,fixed-nodes (apply #'* d)))
;	   (&copy.x ,to ,from (apply #'* d))))
;    (error "move.xx not defined in lisp mode")))


;;; ---------------------------------------------------------------------------
;;;
;;; Squashing
;;;
;;; ---------------------------------------------------------------------------

(defmacro squash (x &optional (f *default-squash-function*))
  (if *c* `(&squash ,x ,f)
    `(lisp-squash ,x ,f)))

(defmacro dsquash (x &optional (f *default-squash-function*))
  (if *c* `(&dsquash ,x ,f)
    `(lisp-dsquash ,x ,f)))

(defmacro squash.v (vec &key squashvec res)
  (if *c*
      `(let* ((n (length ,vec))
	      (svec (or ,squashvec (create-array (length ,vec) :initial-element *default-squash-function*)))
	      (ans (or ,res (create-array n))))
	 (if (or (and ,squashvec
		      (not (= n (length svec))))
		 (and ,res
		      (not (= n (length ,res)))))
	     (error "squash.v -- bad length"))
	 (&squash.x ans ,vec svec n)
	 ans)
    `(squash-vector ,res ,vec ,squashvec)))

(defmacro dsquash.v (vec &key squashvec res)
  (if *c*
      `(let* ((n (length ,vec))
	      (svec (or ,squashvec (create-array (length ,vec) :initial-element *default-squash-function*)))
	      (ans (or ,res (create-array n))))
	 (if (or (and ,squashvec
		      (not (= n (length svec))))
		 (and ,res
		      (not (= n (length ,res)))))
	     (error "dsquash.v -- bad length"))
	 (&dsquash.x ans ,vec svec n)
	 ans)
    `(dsquash-vector ,res ,vec ,squashvec)))


;;; ---------------------------------------------------------------------------
;;;
;;; Equality
;;;
;;; ---------------------------------------------------------------------------

(defmacro =vv (v1 v2 &key (threshold zero))
  (if *c* (macroexpand `(=xx ,v1 ,v2 :threshold ,threshold))
    `(same-vector ,v1 ,v2 :threshold ,threshold)))

(defmacro =mm (m1 m2 &key (threshold zero))
  (if *c* (macroexpand `(=xx ,m1 ,m2 :threshold ,threshold))
    `(same-matrix ,m1 ,m2 :threshold ,threshold)))

(defmacro =xx (x1 x2 &key (threshold zero))
  (if *c*
      `(let ((dims (array-dimensions ,x1)))
	 (unless (equal dims (array-dimensions ,x2))
		 (error "=xx -- bad length"))
	 (zerop (&=xx ,x1 ,x2 (apply #'* dims) ,threshold)))
    (error "=xx not defined for lisp mode")))


;;; ---------------------------------------------------------------------------
;;;
;;; Misc
;;;
;;; ---------------------------------------------------------------------------

(defmacro last.v (v)
  `(aref ,v (1- (length ,v))))

(defmacro cons.v (x vec)
  `(let* ((n (length ,vec))
	  (res (create-array (1+ n))))
     (setf (aref res 0) ,x)
     ,(macroexpand `(move.vv ,vec res :offset 1))   ;unexpanded macro
     res))

(defmacro dist.vv (v1 v2)
  (if *c*
      `(let ((n (length ,v1)))
	 (unless (= n (length ,v2)) (error "dist.vv"))
	 (&dist.vv ,v1 ,v2 n))
    (vector-distance v1 v2)))


;;; ---------------------------------------------------------------------------
;;;
;;; CLOS-like generic function calls
;;;
;;; ---------------------------------------------------------------------------

(defun %* (r s &key res)
  (cond ((and (matrixp r) (matrixp s)) (*mm r s :res res))
	((and (vectorp r) (vectorp s)) (*vv r s :res res))
	((and (matrixp r) (numberp s)) (*ms r s :res res))
	((and (numberp r) (matrixp s)) (*ms s r :res res))
	((and (vectorp r) (numberp s)) (*vs r s :res res))
	((and (numberp r) (vectorp s)) (*vs s r :res res))
	((and (numberp r) (numberp s) (not res) (* r s)))
	(t (error "%*"))))

(defun %+ (r s &key res)
  (cond ((and (matrixp r) (matrixp s)) (+mm r s :res res))
	((and (vectorp r) (vectorp s)) (+vv r s :res res))
	((and (matrixp r) (numberp s)) (+ms r s :res res))
	((and (numberp r) (matrixp s)) (+ms s r :res res))
	((and (vectorp r) (numberp s)) (+vs r s :res res))
	((and (numberp r) (vectorp s)) (+vs s r :res res))
	((and (numberp r) (numberp s) (not res) (+ r s)))
	(t (error "%+"))))

(defun %. (r s &key res)
  (cond ((and (matrixp r) (matrixp s))           (.mm r s :res res))
	((and (vectorp r) (vectorp s) (not res)) (.vv r s))
	((and (matrixp r) (vectorp s))           (.mv r s :res res))
	((and (vectorp r) (matrixp s))           (.vm r s :res res))
	(t (error "%."))))
