;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;
;;;     			     ANSIL
;;; 		      Advanced Network Simulator In Lisp
;;; 		      ----------------------------------
;;;
;;;				  Written By
;;;			       Peter J. Angeline
;;;			      Gregory M. Saunders
;;;
;;;			   The Ohio State University
;;;		       Laboratory for AI Research (LAIR)
;;;			     Columbus, Ohio, 43210
;;;
;;;			   Copyright (c) 1991, 1992
;;;
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; ---------------------------------------------------------------------- 
;;; File: net1.lisp - This file contains a matrix implementation of networks.
;;; This is just the basic setup routines.  For more user-friendly stuff, see
;;; net2.lisp.
;;;
;;; Author: Gregory M Saunders
;;;
;;; License: This code may be distributed free of charge.  This code may not be
;;; incorporated into any production software of any type or any code which is
;;; sold without the express written consent of the authors.  Modifications and
;;; inclusions for other types of software are permitted, provided that this
;;; header remains unchanged and included in all subsequent versions of the
;;; code.
;;;
;;; Disclaimer: The authors of this code make no claims as to its
;;; validity, and are in no way responsible for any errors contained within.
;;; 
;;; File description: This file contains just the bare bones:  network
;;; structure, forward-pass, backward-pass.  Names follow the PDP conventions,
;;; within the limits of lower-case English languages.
;;;
;;; Creation date: March 17, 1992, Version 2.01
;;; ----------------------------------------------------------------------
;;;
;;;
;;;                         Change Log
;;;                         ----------
;;;
;;; ----------------------------------------------------------------------

;;; ---------------------------------------------------------------------------
;;;
;;; Variables
;;;
;;; ---------------------------------------------------------------------------

(defvar *default-rate*     .7d0)
(defvar *default-momentum* .5d0)
(defvar *default-epsilon*  .1d0)
(defvar *default-sp-add*   .1d0)
(defvar *default-bias*      1d0)


;;; ----------------------------------------------------------------------
;;;
;;; Network structure
;;;
;;; ----------------------------------------------------------------------

;; ASSUME N=Number of nodes
;;    ===> N^2 weights

(deftype matrix  () '(array double-float))
(deftype dvector () '(simple-array double-float))

(defstruct network
  "Everything you ever wanted in a network, and then some."
  (maxnode           0 :type integer)		; number of nodes in network - 1
  (total-epochs      0 :type integer)		; total number of passes through training set
  (weights         nil :type matrix)		; N^2 matrix
  (dweights        nil :type matrix)		; N^2 matrix = DELTAS
  (pdweights       nil :type matrix)		; N^2 matrix = previous DELATS
  (free-weights    nil :type matrix)		; N^2 matrix that tells which weights are not fixed
  (activations     nil :type dvector)		; N vector
  (deltas          nil :type dvector)		; N vector
  (fclamp          nil :type dvector)		; N vector - clamped nodes for forward pass
  (bclamp          nil :type dvector)		; N vector - clamped nodes for backward pass
  (squashvec       nil :type dvector)		; N vector, 0=constant, 1=sigmoid, 2=tanh
  (adjust?         nil :type symbol)		; adjust rate by pattern number?
  (initial-weights nil :type vector)		; random seed which gives init weights
  (initial-max    .5d0 :type double-float)	; max value of random weights for this net
  (rate            0d0 :type double-float)	; learning rate
  (momentum        0d0 :type double-float)	; momentum term
  (epsilon         0d0 :type double-float)	; acceptable error distance 
  (max-error       0d0 :type double-float)	; maximum error found in last bp  
  (total-error     0d0 :type double-float)	; total error from last bp 
  (bias            0d0 :type double-float)	; the value for bias activation
  (sp-add          0d0 :type double-float)	; constant to add to sigmoid derivative
  (bias-node       nil :type symbol)		; if t, cons the bias onto network before fpass
  (always-bp?        t :type symbol)		; bp or not when within tolerence -- unsupported
  (tset            nil :type t)			; the training set for this network, type unspecified
  ;; (pushprop-patterns nil :type list)         ; the patterns for pushprop assocs
  )


;;; ----------------------------------------------------------------------
;;;
;;; Generic network stuff
;;;
;;; ----------------------------------------------------------------------

(defun generic-network (n		;NUMBER OF NODES IN NETWORK
			&key
			rate
			momentum
			epsilon
			sp-add
			bias
			(bias-node nil)
			(adjust? nil)
			(always-bp? t)
			(squashvec (create-array n :initial-element 1))
			(fclamp    (create-array n))
			(bclamp    (create-array n))
			&aux
			(net (make-network))
			(dims (list n n)))
  (setf (network-maxnode net) (1- n))
  (setf (network-weights net) (create-array dims))
  (set-network net 
	       :rate       (or rate     *default-rate*)
	       :momentum   (or momentum *default-momentum*)
	       :epsilon    (or epsilon  *default-epsilon*)
	       :sp-add     (or sp-add   *default-sp-add*)
	       :bias       (or bias     *default-bias*)
	       :bias-node  bias-node
	       :adjust?    adjust?
	       :always-bp? always-bp?
	       :squashvec  squashvec
	       :fclamp     fclamp
	       :bclamp     bclamp)
  net)

;; function to set all of the setable things in a network.
(defun set-network (net &key (rate       (network-rate       net))
			     (momentum   (network-momentum   net))
			     (epsilon    (network-epsilon    net))
			     (bias       (network-bias       net))
			     (bias-node  (network-bias-node  net))
			     (sp-add     (network-sp-add     net))
			     (adjust?    (network-adjust?    net))
			     (always-bp? (network-always-bp? net))
			     (squashvec  (network-squashvec  net))
			     (fclamp     (network-fclamp     net))
			     (bclamp     (network-bclamp     net)))
  (unless always-bp? (error "set-network -- \"always-bp? = nil\" not yet supported"))
  (unless (= (1+ (network-maxnode net))
	     (length squashvec)
	     (length fclamp)
	     (length bclamp))
	  (error "set-network -- bad length"))
  (setf (network-rate       net) (coerce-number rate))
  (setf (network-momentum   net) (coerce-number momentum))
  (setf (network-epsilon    net) (coerce-number epsilon))
  (setf (network-bias	    net) (coerce-number bias))
  (setf (network-sp-add	    net) (coerce-number sp-add))
  (setf (network-squashvec  net) (coerce-vector squashvec))
  (setf (network-fclamp     net) (coerce-vector fclamp))
  (setf (network-bclamp     net) (coerce-vector bclamp))
  (setf (network-adjust?    net) adjust?)
  (setf (network-always-bp? net) always-bp?)
  (setf (network-bias-node  net) bias-node)
  (format nil "~%Network Properties:~%~{   ~28a: ~a~%~}~%"
	  `( ;"Element Type"                 ,*element-type*
	    "Learning Rate (rate)"	   ,(network-rate net)
	    "Momentum (momentum)"	   ,(network-momentum net)
	    "Tolerance (epsilon)"	   ,(network-epsilon net)
	    "Bias Activation (bias)"       ,(network-bias net)
	    "Bias-node"                    ,(network-bias-node net)
	    "Adjust Rate? (adjust?)"       ,(network-adjust? net)
	    "Always BP? (always-bp?)"      ,(network-always-bp? net)
	    "Sigmoid Prime Add (sp-add)"   ,(network-sp-add net)
	    "Squashvec"		           ,(format-vector (network-squashvec net))
	    "Fpass clamped nodes (fclamp)" ,(format-vector (network-fclamp net))
	    "Bpass clamped nodes (bclamp)" ,(format-vector (network-bclamp net)))))


(defun zero-out-network (net &aux (n (1+ (network-maxnode net)))
			          (dims (list n n)))
  (setf (network-dweights net)  (create-array dims))
  (setf (network-pdweights net) (create-array dims))
  (setf (network-activations net) (create-array n))
  (setf (network-deltas net)      (create-array n)))

(defun clear-deltas (net &aux (deltas (network-deltas net)))
  (loop for i from 0 to (network-maxnode net) do
	(setf (aref deltas i) zero)))


;;; ---------------------------------------------------------------------------
;;;
;;; Setting free-weights
;;;
;;; ---------------------------------------------------------------------------

(defun connect-nodes (net n1 n2)
  (set-connection net n1 n2 one))

(defun disconnect-nodes (net n1 n2)
  (set-connection net n1 n2 zero))

(defun set-connection (net n1 n2 val)
  (cond ((and (numberp n1) (numberp n2)) (setf (aref (network-free-weights net) n1 n2) val))
	((and (listp n1)   (numberp n2)) (loop for i in n1 do (set-connection net i n2 val)))
	((and (numberp n1) (listp n2))   (loop for j in n2 do (set-connection net n1 j val)))
	(t (loop for i in n1 do
		 (loop for j in n2 do
		       (set-connection net i j val))))))

(defmacro free-weight (net i j)
  `(connect-nodes ,net ,i ,j))

(defun set-weight (net i j &optional val)
  (if val (setf (aref (network-weights net) i j) (coerce-number val))
    (aref (network-weights net) i j)))

(defun fix-weight (net i j &optional val)
  (disconnect-nodes net i j)
  (if val (set-weight net i j val)))



;;; ----------------------------------------------------------------------
;;;
;;; Random routines
;;;
;;; ----------------------------------------------------------------------
;; randomize the network and store the seed for the initial conditions
(defun randomize-network (net &optional (max 0.5))
  (setf *random-seed* (ansil-make-random-state))
  (setf (network-initial-weights net) (copy-seq *random-seed*))
  (setf (network-initial-max net) max)
  (revert-network net))

;; revert to the initial weghts before training
(defun revert-network (net)
  (setf *random-seed* (copy-seq (network-initial-weights net)))
  (randomize-matrix (network-weights net)
		    :max (network-initial-max net)
		    :free-weights (network-free-weights net))
  (zero-out-network net)
  (setf (network-total-epochs net) 0)
  nil)

(defun randomize-matrix (matrix &key (max .5) (free-weights 'all))
  (loop for i from 0 to (- (array-dimension matrix 0) 1) do
	(loop for j from 0 to (- (array-dimension matrix 1) 1)
	      do
	      (if (or (equal free-weights 'all)
		      (= (aref free-weights i j) one))
		  (setf (aref matrix i j)
			(coerce (random-number max) *element-type*))))))

(defun randomize-vector (vector &optional (max .5))
  (loop for i from 0 to (- (length vector) 1)
	do (setf (aref vector i) (coerce (random-number max) *element-type*))))

(defun random-number (n)
  (* (if (zerop (random-int 2)) 1.0 -1.0) (random-float n)))


;;; ---------------------------------------------------------------------------
;;;
;;; Forward pass 
;;;   NOTE:  this returns a POINTER to network-activations.
;;;
;;; ---------------------------------------------------------------------------

(defun fpass (net vec
		  &key
		  (num -1)
		  (debug nil)
		  (squashvec (network-squashvec net))
		  (bias-node (network-bias-node net))
		  (fixed-nodes (network-fclamp net))
		  (threshold .01d0)
		  &aux
		  (x    (network-activations net))
		  (oldx (create-array (1+ (network-maxnode net)) :initial-element -999))
		  (y    (create-array (1+ (network-maxnode net))))
		  (w    (network-weights net)))
  (unless (vectorp vec) (setf vec (coerce-vector vec)))
  (if bias-node (setf vec (cons.v (network-bias net) vec)))
  (move.vv vec x)
  (loop until (or (zerop num)
		  (and (< num 0) (=vv x oldx :threshold threshold)))
	do
	(when debug (format t "activations = ~a~%" x))
	(decf num)
	(move.vv x oldx)
	(.vm x w :res y)
	(squash.v y :squashvec squashvec :res y)
	(move.vv y x :fixed-nodes fixed-nodes))
  x)



;;; ---------------------------------------------------------------------------
;;;
;;; Backward pass
;;;   NOTE:  bpass returns a POINTER to network-deltas.
;;;
;;; ---------------------------------------------------------------------------

;;; see PDP, p. 327
;;;
;;;  always-bp?
;;;  adjust-rate?

;; output unit deltas
(defun set-deltas (net &key target
		       &aux
		       (ndeltas (network-deltas net))
		       (activations (network-activations net))
		       (squashvec (network-squashvec net))
		       (offset (1+ (- (network-maxnode net) (length target)))))
  (unless (vectorp target) (setf target (coerce-vector target)))
  (loop for i from 0 to (1- (length target))
	for j = (+ i offset) 
	for o = (aref activations j)
	do
	;;	(format t "i=~a, o=~a, sv=~a, s=~a~%" i o (aref squashvec j) (dsquash o (aref squashvec j)))
	(setf (aref ndeltas j) (- (aref target i) o)))
  (*vv ndeltas 
       (+vs (dsquash.v activations :squashvec squashvec) (network-sp-add net)) ;??? sp-add
       :res ndeltas))

(defun bpass (net &key
		  target 
		  deltas
		  (num -1)
		  (debug nil)
		  (squashvec (network-squashvec net))
		  (fixed-nodes (network-bclamp net))
		  (threshold .01d0)
		  &aux
		  (x    (network-deltas net))
		  (oldx (create-array (1+ (network-maxnode net)) :initial-element -999))
		  (y    (create-array (1+ (network-maxnode net))))
		  (dact (+vs (dsquash.v (network-activations net) :squashvec squashvec)
			     (network-sp-add net))) ;???
		  (w    (network-weights net)))
  (cond ((and target deltas) (error "can't specify both target and deltas"))
	(target (clear-deltas net)
		(set-deltas net :target target))
	(deltas (if (not (vectorp deltas)) (error "bad deltas"))
		(clear-deltas net)
		(move.vv deltas x :offset (- (length x) (length deltas))))
	(t (error "must specify either target or deltas")))
  (loop until (or (zerop num)
		  (and (< num 0) (=vv x oldx :threshold threshold)))
	do
	(decf num)
	(when debug (format t "deltas = ~a~%" x))
	(move.vv x oldx)
	(.mv w x :res y)
	(*vv y dact :res y)
	(move.vv y x :fixed-nodes fixed-nodes))
  x)
