;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;
;;;     			     ANSIL
;;; 		      Advanced Network Simulator In Lisp
;;; 		      ----------------------------------
;;;
;;;				  Written By
;;;			       Peter J. Angeline
;;;			      Gregory M. Saunders
;;;
;;;			   The Ohio State University
;;;		       Laboratory for AI Research (LAIR)
;;;			     Columbus, Ohio, 43210
;;;
;;;			   Copyright (c) 1991, 1992
;;;
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; ---------------------------------------------------------------------- 
;;; File: net2.lisp - Sample network code built on top of net1.lisp.  For
;;; instance, backprop routines etc go here.
;;;
;;; Author: Gregory M Saunders
;;;
;;; License: This code may be distributed free of charge.  This code may not be
;;; incorporated into any production software of any type or any code which is
;;; sold without the express written consent of the authors.  Modifications and
;;; inclusions for other types of software are permitted, provided that this
;;; header remains unchanged and included in all subsequent versions of the
;;; code.
;;;
;;; Disclaimer: The authors of this code make no claims as to its
;;; validity, and are in no way responsible for any errors contained within.
;;; 
;;; File description:  This file builds backprop on top of the bare bones
;;; provided by net1.lisp.
;;;
;;; Creation date: March 13, 1992, Version 2.01
;;; ----------------------------------------------------------------------
;;;
;;;
;;;                         Change Log
;;;                         ----------
;;;
;;; ----------------------------------------------------------------------

(defun create-network (num-nodes &rest parms &key &allow-other-keys
		       (type 'feed-forward)
		       (layers nil)
		       (initial-max .5)
		       (bias-node (eql type 'feed-forward))
		       (fixed-weights nil)
		       (debug nil)
		       &aux
		       (n (if bias-node (1+ num-nodes) num-nodes))
		       (net (sapply #'generic-network n (remove-keywords parms
							 :type :debug :layers :initial-max
							 :fixed-weights)))
		       (dims (list n n)))
  (if (and layers (not (eql type 'feed-forward))) (error "layers is for feed-forward nets only"))
  (if bias-node (set-network net :bias-node t))
  (case type
	(hopfield 
	 (setf (network-free-weights net) (create-array dims :initial-element 1))
	 (setf (network-squashvec net)    (create-array n    :initial-element 0)))
	(sparse   (setf (network-free-weights net) (create-array dims :initial-element 0)))
	(dense    (setf (network-free-weights net) (create-array dims :initial-element 1)))
	(feed-forward
	 (progn
	   (unless layers (error "Must specify layers for feed-forward net"))
	   (setf (network-free-weights net) (create-array dims :initial-element 0))
;	   (loop for i from 0 to (if bias-node (car layers) (1- (car layers))) do
;		 (fix-weight net i i 1))
	   (if bias-node (loop for i from (1+ (car layers)) to (1- n) do (free-weight net 0 i)))
;	   (loop for i from 0 to (car layers) do (setf (aref (network-squashvec net) i) zero))
	   (unless (= (1- n)
		      (loop for num1 in layers
			    for num2 in (cdr layers)
			    for x1 = (if bias-node 1 0) then (1+ x2)
			    for x2 = (1- (+ x1 num1))
			    for y1 = (+ x1 num1)
			    for y2 = (1- (+ y1 num2))
			    do
			    (if (= x1 (if bias-node 1 0))
				(loop for i from 0 to x2 do (setf (aref (network-fclamp net) i) 1d0)))
			    (if (= y2 (1- n))
				(loop for i from y1 to y2 do (setf (aref (network-bclamp net) i) 1d0)))
			    (if (>= y2 n) (error "layer specification"))
			    (connect-nodes net
					   (loop for i from x1 to x2 collect i)
					   (loop for j from y1 to y2 collect j))
			    finally (return y2)))
		   (format t "Warning:  level specification left out some nodes~%"))))
	(otherwise (error "unrecognized type")))
  (loop for x in fixed-weights 
	for i = (car x)
	for j = (cadr x)
	do
	(fix-weight net i j (caddr x)))
  (randomize-network net initial-max)
  (if debug (show-network net))
  net)



;;; ---------------------------------------------------------------------------
;;;
;;; BACKPROP STUFF
;;;
;;; ---------------------------------------------------------------------------

;;; ---------------------------------------------------------------------------
;;;
;;; Updating
;;;
;;; ---------------------------------------------------------------------------

;; I removed the rate multiplication from this step to speed things up a bit.
;; Instead it's in update-weights.  The problem is that we loose the ability to
;; do repetitions, but this is just generic backprop.

(defun compute-dweights (net &key
;			     (divisor one) (repetitions one)
			     &aux 
;			     (rate (* (network-rate net) (/ repetitions divisor)))
			     (dweights (outer (network-activations net)
					      (network-deltas net))))
  (+mm dweights (network-dweights net) :res (network-dweights net)))


;;; ---------------------------------------------------------------------------
;;;
;;; Update weights
;;;
;;;    dW <--- a dW + b pdW
;;;     W <--- W + dW
;;;   pdW <--- dW
;;;    dW <--- 0
;;;
;;; ---------------------------------------------------------------------------

(defun update-weights (net &aux
			   (w   (network-weights net))
			   (dw  (network-dweights net))
			   (pdw (network-pdweights net))
			   (rate (network-rate net))
			   (momentum (network-momentum net))
			   (tmp pdw))
  (when (network-adjust? net)
	(setf rate (/ rate (length (network-tset net)))))
  (*ms dw rate :res dw)
  (*ms pdw momentum :res pdw)
  (+mm dw pdw :res dw)
  (*mm dw (network-free-weights net) :res dw)
  (+mm w dw :res w)
  (setf (network-pdweights net) dw)
  (setf (network-dweights net) tmp)     ;; zeroing out the old pdw and using it
  (zero-matrix (network-dweights net))) ;; for the new dw saves space - from pja

;(defun update-pdweights (net &aux (temp (network-dweights net)))
;  (setf (network-dweights net) (network-pdweights net))
;  (setf (network-pdweights net) temp)
;  (zero-matrix (network-dweights net)))

;; make all the elements in the matrix to be 0.0
(defun zero-matrix (m)
  (*ms m zero :res m))

;;; ---------------------------------------------------------------------------
;;;
;;; Manipulating the training set
;;;
;;; ---------------------------------------------------------------------------

(defun add-pattern (net in out &aux (n (1+ (network-maxnode net))))
  (if (or (> (length in) n)
	  (> (length out) n))
      (error "add-pattern -- pattern too long"))
  (setf (network-tset net)
	(reverse
	 (cons (list (coerce-vector in) (coerce-vector out))
	       (reverse (network-tset net))))))


;;; ---------------------------------------------------------------------------
;;;
;;; Training 
;;;
;;; ---------------------------------------------------------------------------

(defun network-train (net &key
			  (num 10000)
			  (display 1) 
			  (check-every display)
			  (tset (network-tset net))
			  &aux (errors 1))
  (unless (typep net 'network) (error "network expected"))
  (zero-matrix (network-dweights net))
  (zero-matrix (network-pdweights net))
  (loop for count    = 1 then (1+ count)
	for displayp = (zerop (mod count display))
	for checkp   = (or (zerop (mod count check-every))
			   (= count num))
	until (or (> count num)
		  (and (zerop (mod (1- count) check-every))
		       (zerop errors)))
	do
	(setf (network-total-error net) zero
	      (network-max-error net) zero)
	(setf errors (network-train-on-pairs net tset (or checkp displayp) displayp))
	(incf (network-total-epochs net))
	(when displayp (show-network-training-stats net errors))
	finally 
	(progn (format t "~%")
	       (show-network-training-stats net errors)))
  (zerop errors))

(defun show-network-training-stats (net errors)
  (format t "Epoch ~d: Err= ~d Tot=~,6f Max=~,6f~%"  ;format
	  (network-total-epochs net)
	  errors (network-total-error net)
	  (network-max-error net))
  (force-output *standard-output*))

(defun network-train-on-pairs (net pairs checkp displayp &aux (errors 0))
  (loop for pair in pairs
;	for input  = (car pair)
;	for target = (cadr pair)
	do
	(incf errors (network-train-on-pair net (car pair) (cadr pair) checkp displayp)))
  (if displayp (format t "~%"))            ;format
  (update-weights net)
  errors)

(defun network-train-on-pair (net input target checkp displayp)
  (fpass net input)
  (bpass net :target target)
  (compute-dweights net)
  (if checkp (count-errors net target :display displayp) 1))

(defun count-errors (net target &key display 
			 &aux
			 (epsilon (network-epsilon net))
			 (activations (network-activations net))
			 (offset (1+ (- (network-maxnode net) (length target)))))
  (loop for i from 0 to (1- (length target))
	for j   = (+ i offset)
	for err = (abs (- (aref target i) (aref activations j)))
	sum (* err err) into total-error
	maximize err into max-error
	count (> err epsilon) into num
	finally
	(when display (write-char (if (zerop num) #\. #\*)))
	(incf (network-total-error net) total-error)
	(setf (network-max-error net)
	      (max max-error (network-max-error net)))
	(return num)))


;;; ---------------------------------------------------------------------------
;;;
;;; I/O
;;;
;;; ---------------------------------------------------------------------------

(defun write-network (network filename)
  (unless (typep network 'network) (error "write network"))
  (let ((acts      (network-activations network))
	(deltas    (network-deltas network))
	(dweights  (network-dweights network))
	(pdweights (network-pdweights network)))
    (setf (network-activations network) nil      ;don't save this garbage...
	  (network-deltas      network) nil	
	  (network-dweights    network) nil
	  (network-pdweights   network) nil)
    (with-open-file (fi (concatenate 'string filename ".network")
			:direction :output :if-exists :supersede)
		    (pprint network fi))
    (setf (network-activations network) acts     ; ...but be sure to restore it
	  (network-deltas      network) deltas 
	  (network-dweights    network) dweights
	  (network-pdweights   network) pdweights))
  nil)

(defun read-network (filename)
  (let ((net (with-open-file (fi (concatenate 'string filename ".network")
				 :direction :input)
			     (read fi))))
    (zero-out-network net)              ;make sure we've got arrays, not nils 
    (fix-network-arrays net)            ;and make sure everything is double-float
    net))



;;; ----------------------------------------------------------------------
;;;
;;; Ugly patch to fix a problem with saving/restoring networks.
;;; The problem is that when an array of type DOUBLE-FLOAT is written to
;;; disk, it will be read back in as an array of type TRUE.  The following
;;; routine alleviates this problem for networks by walking through the
;;; structure and effectively changing all arrays to be of type
;;; *element-type*.  Yuck.  
;;;
;;; See raam.lisp for a similar fix for raams.
;;;
;;; ----------------------------------------------------------------------

(defun fix-network-arrays (net)
  (unless (equal *element-type*
		 (type-of (network-rate net)))
	  (error (format nil "desired network not of type ~a" *element-type*)))

  (setf (network-weights         net) (fix-array (network-weights         net)))
  (setf (network-free-weights    net) (fix-array (network-free-weights    net)))
  (setf (network-fclamp          net) (fix-array (network-fclamp          net)))
  (setf (network-bclamp          net) (fix-array (network-bclamp          net)))
  (setf (network-squashvec       net) (fix-array (network-squashvec       net)))
  (setf (network-initial-weights net) (fix-array (network-initial-weights net) :element-type 'fixnum))
  (fix-network-tset net))

(defun array->list (m)
  (let ((dim (array-dimensions m)))
    (cond ((= (length dim) 1)
	   (loop for i from 0 to (1- (car dim)) collect (aref m i)))
	  ((= (length dim) 2)
	   (loop for i from 0 to (1- (car dim)) collect
		 (loop for j from 0 to (1- (cadr dim)) collect
		       (aref m i j))))
	  (t (error "array->list supports only arrays of size <= 2")))))

(defun fix-array (bad-array &key (element-type *element-type*))
  (make-array (array-dimensions bad-array)
	      :element-type element-type
	      :initial-contents (array->list bad-array)))

(defun fix-network-tset (net)
  (setf (network-tset net)
	(loop for x in (network-tset net)
	      collect
	      (list (fix-array (car x)) (fix-array (cadr x))))))


;;; ---------------------------------------------------------------------------
;;;
;;; A couple of useful functions to get the input/output nodes of a network
;;;
;;; ---------------------------------------------------------------------------

(defun network-input-size (net)
  (- (loop for x across (network-fclamp net)
	   for i = 0 then (1+ i)
	   when
	   (zerop x)
	   return i)
     (if (network-bias-node net) 1 0)))

(defun network-output-size (net &aux (bclamp (network-bclamp net)))
  (loop for i from (1- (length bclamp)) downto 0
	for x = (aref bclamp i)
	when
	(zerop x)
	return (- (length bclamp) i 1)))

(defun network-inputs (net &optional (vec (network-activations net)))
  (format-vector
   (extract vec :end (network-input-size net))))

(defun network-outputs (net &optional (vec (network-activations net)))
  (format-vector 
   (extract vec :start (- (network-maxnode net)
			  (network-output-size net)
			  -1))))
