;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;
;;;     			     ANSIL
;;; 		      Advanced Network Simulator In Lisp
;;; 		      ----------------------------------
;;;
;;;				  Written By
;;;			       Peter J. Angeline
;;;			      Gregory M. Saunders
;;;
;;;			   The Ohio State University
;;;		       Laboratory for AI Research (LAIR)
;;;			     Columbus, Ohio, 43210
;;;
;;;			   Copyright (c) 1991, 1992
;;;
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; ---------------------------------------------------------------------- 
;;; File: raam.lisp - Basic raam functionality.
;;;
;;; Author: Gregory M Saunders
;;;
;;; License: This code may be distributed free of charge.  This code may not be
;;; incorporated into any production software of any type or any code which is
;;; sold without the express written consent of the authors.  Modifications and
;;; inclusions for other types of software are permitted, provided that this
;;; header remains unchanged and included in all subsequent versions of the
;;; code.
;;;
;;; Disclaimer: The authors of this code make no claims as to its
;;; validity, and are in no way responsible for any errors contained within.
;;; 
;;; File description: This file contains an implementation raams.  70% of the
;;; functions are just conveniences for analysis.
;;;
;;; Creation date:  April 1, 1992, Version 2.02
;;; ----------------------------------------------------------------------
;;;
;;;
;;;                         Change Log
;;;                         ----------
;;;
;;; ----------------------------------------------------------------------


;;; ---------------------------------------------------------------------------
;;;
;;; Structures
;;;
;;; ---------------------------------------------------------------------------

(defstruct raam
  (encoder nil :type network)
  (decoder nil :type network)
  (termnet nil :type network)
  (symtab0 nil :type list)
  (symtab1 nil :type list)
  (tset    nil :type list)
  (termtest  0 :type integer)  ; 0 for terminal bits, 1 for give structure to raam
  (repwidth  0 :type integer)
  (valence   0 :type integer))

(defun create-encoder (n1 n2 &rest parms &key &allow-other-keys)
  (sapply #'create-network
	  (+ n1 n2)
	  :type 'feed-forward
	  :layers (list n1 n2)
	  parms))
				  
(defun create-decoder (n1 n2 &rest parms &key &allow-other-keys)
  (sapply #'create-encoder n1 n2 parms))

(defun create-raam (&rest parms &key &allow-other-keys
		    tset
		    symtab
		    valence
		    repwidth
		    (termtest 0)
		    &aux
		    (args (remove-keywords parms :tset :symtab :valence :repwidth :termtest))
		    (raam (make-raam 
			   :encoder (sapply #'create-encoder (* repwidth valence) repwidth args)
			   :decoder (sapply #'create-decoder repwidth (* repwidth valence) args)
			   :termnet (sapply #'create-network 
					    (+ repwidth valence)
					    :type 'feed-forward
					    :layers (list repwidth valence)
					    args)
			   :tset tset
			   :termtest termtest)))
  (setf (raam-symtab0 raam)
	(loop for xv in symtab 
	      for x = (car xv)
	      for v = (cadr xv)
	      collect
	      (list x (coerce-vector v))))
  (setf (raam-repwidth raam) repwidth)
  (setf (raam-valence raam)   valence)
;  (when sp-add
;	(set-network (raam-encoder raam) :sp-add sp-add)
;	(set-network (raam-decoder raam) :sp-add sp-add))
  raam)


;;; ---------------------------------------------------------------------------
;;;
;;; Encoding
;;;
;;; ---------------------------------------------------------------------------

(defun lookup (raam sym)
  (cadr
   (or (assoc sym (raam-symtab0 raam))
       (assoc sym (raam-symtab1 raam) :test #'equal))))

(defun memorize-tree (raam tree vec)
  (setf (raam-symtab1 raam)
	(cons (list tree vec) (raam-symtab1 raam))))

(defun forget-trees (raam)
  (setf (raam-symtab1 raam) nil))

(defun encode (raam tree)
  (or (lookup raam tree)
      (let* ((repwidth (raam-repwidth raam))
	     (valence  (raam-valence  raam))
	     (in (apply #'append.vecs (loop for subtree in tree collect (encode raam subtree)))))
	(if (or (atom tree) 
		(not (= (length tree) valence)))
	    (error "bad tree"))
	(let ((out (extract (fpass (raam-encoder raam) in)
			    :start (1+ (* repwidth valence)))))
					;	  (format t "Old tree = ~a~%" in)
					;	  (format t "New tree = ~a~%" out)
	  (memorize-tree raam tree out)
	  out))))

;;; ---------------------------------------------------------------------------
;;;
;;; Decoding
;;;
;;; ---------------------------------------------------------------------------

(defun symbol-name-by-value (raam vec &key (terminals-only nil))
  (let* ((symtab (append (raam-symtab0 raam)
			 (if terminals-only nil (raam-symtab1 raam))))
	 (distances (loop for sv in symtab collect
			  (dist.vv (cadr sv) vec)))
	 (min (apply #'min distances))
	 (index (position min distances)))
    (values (car (nth index symtab))
	    min)))

(defun decode (raam vec &key parse &aux (repwidth (raam-repwidth raam)) res)
  (unless (= (length vec) repwidth) (error "bad vector"))
  (setf res (extract (fpass (raam-decoder raam) vec) :start (1+ repwidth)))
  (values (if parse (parse raam res) res)
	  (extract (fpass (raam-termnet raam) vec)
		   :start (1+ repwidth))))

(defun parse (raam vec
		   &aux 
		   (valence (raam-valence raam))
		   (repwidth (raam-repwidth raam)))
  (loop for i from 0 to (1- valence) collect
	(extract vec
		 :start (* i repwidth)
		 :end   (1- (* (1+ i) repwidth)))))

(defun rdecode (raam vec &key target (max-depth 10) (debug nil))
;  (if (or (eq target t)
;	  (and (= (raam-termtest raam) 1)
;	       (not target)))
;      (setf target vec))        ;; not supported yet, since it's a vec, not a tree
  (multiple-value-bind
   (termvecs termbits)
   (decode raam vec :parse t)
   (loop for outbit across termbits
	 for outvec in termvecs
	 for i = 0 then (1+ i)
	 for tree = (nth i target)
					;	 i from 0 to (1- valence)
					;	 for outbit = (nth i termbits)
					;	 for outvec = (nth i termvecs)
	 collect
	 (progn 
	   (if debug (print-for-debugging raam outbit outvec))
	   (if (terminal-test raam outvec outbit max-depth :target tree)
	       (terminal-vector-output raam outvec max-depth)
	     (prog1
	       (rdecode raam outvec 
			:debug debug
			:target tree
			:max-depth (1- max-depth))
	       (if debug (format t "~%"))))))))


;;; ---------------------------------------------------------------------------
;;;
;;; Terminal test, debugging, output from a terminal-vector.  These are the
;;; routines that you should set if you don't like either of the two terminal
;;; tests provided.
;;;
;;; ---------------------------------------------------------------------------

(defun terminal-test (raam vector termbit max-depth &key target)
  (declare (ignore vector))
  (case (raam-termtest raam)
	(0 (or (zerop max-depth) 
	       (> termbit .5d0)))
	(1 (terminal target))
	(t (error "terminal-test"))))

(defun print-for-debugging (raam termbit vec)
  (multiple-value-bind (symbol dist)                      
		       (symbol-name-by-value raam vec :terminals-only t)
		       (format t "  ~5,2f * ~a ===> ~a, dist = ~4,2f~%"
			       termbit
			       (format-vector vec)
			       symbol
			       dist)))

(defun terminal-vector-output (raam vec max-depth)
  (if (zerop max-depth) '*
    (symbol-name-by-value raam vec :terminals-only t)))




;;; ---------------------------------------------------------------------------
;;;
;;; Training
;;;
;;; ---------------------------------------------------------------------------

(defun terminal (x)
  (if (null x) (error "terminal")
    (atom x)))

(defun raam-train-on-tree (raam tree checkp displayp &aux deltas deltas2)
  (let* ((y (encode raam tree))
	 (decoder (raam-decoder raam))
	 (encoder (raam-encoder raam))
	 (termnet (raam-termnet raam))
	 (repwidth (raam-repwidth raam))
	 (target-vector (apply #'append.vecs (loop for s in tree collect (lookup raam s))))
	 (target-bits   (coerce-vector (loop for s in tree collect (if (terminal s) 1d0 0d0)))))
    (decode raam y)
    (bpass decoder :target target-vector)
    (bpass termnet :target target-bits)
    (setf deltas (extract (network-deltas decoder) :start 1 :end repwidth))
    (when t				;(raam-train-terminal-bits raam)
	  (setf deltas2 (extract (network-deltas termnet) :start 1 :end repwidth))
	  (+vv deltas deltas2 :res deltas))
    (bpass encoder :deltas deltas)
    (compute-dweights decoder)
    (compute-dweights encoder)
    (compute-dweights termnet)
    (if checkp 
	(+ (count-errors decoder target-vector :display displayp)
	   (count-errors termnet target-bits   :display displayp))
      1)))

(defun raam-train-on-trees (raam trees checkp displayp &aux (errors 0))
  (forget-trees raam)
  (loop for tree in trees
	do
	(incf errors (raam-train-on-tree raam tree checkp displayp)))
  (if displayp (format t "~%"))
  (update-weights (raam-encoder raam))
  (update-weights (raam-decoder raam))
  (update-weights (raam-termnet raam))
  errors)

(defun raam-train (raam
		   &key
		   (num 10000)
		   (display 1) 
		   (check-every display)
		   (tset (if (raam-p raam) (raam-tset raam)))
		   &aux
		   (errors 1)
		   (encoder (raam-encoder raam))
		   (decoder (raam-decoder raam))
		   (termnet (raam-termnet raam)))
  (unless (raam-p raam) (error "raam-train given non-raam"))
  (zero-matrix (network-dweights encoder))
  (zero-matrix (network-pdweights encoder))
  (zero-matrix (network-dweights decoder))
  (zero-matrix (network-pdweights decoder))
  (zero-matrix (network-dweights termnet))
  (zero-matrix (network-pdweights termnet))
  (loop for count    = 1 then (1+ count)
	for displayp = (zerop (mod count display))
	for checkp   = (or (zerop (mod count check-every))
			   (= count num))
	until (or (> count num)
		  (and (zerop (mod (1- count) check-every))
		       (zerop errors)))
	do
	;;	(setf (network-total-error encoder) zero
	;;	      (network-max-error encoder) zero)
	(setf (network-total-error decoder) zero
	      (network-max-error decoder) zero)
	(setf (network-total-error termnet) zero
	      (network-max-error termnet) zero)
	(setf errors (raam-train-on-trees raam tset (or checkp displayp) displayp))
	;;	(incf (network-total-epochs encoder))
	(incf (network-total-epochs decoder))
	(incf (network-total-epochs termnet))
	(when displayp 
	      ;; (show-network-training-stats encoder errors)
	      (show-network-training-stats decoder errors)
	      (show-network-training-stats termnet errors))
	finally 
	(progn
	  ;; (show-network-training-stats encoder errors)
	  (format t "~%")
	  (show-network-training-stats decoder errors)
	  (show-network-training-stats termnet errors)))
  (zerop errors))


;;; ---------------------------------------------------------------------------
;;;
;;; Show
;;;
;;; ---------------------------------------------------------------------------

(defun max-length (r &key (key #'identity))
  (if (null r) 0
    (apply #'max (loop for x in r collect
		       (length (to-string (funcall key x)))))))

(defun to-string (x &optional (n nil))
  (let ((temp$ (princ-to-string x)))
    (if n (subseq temp$ 0 (min n (length temp$)))
      temp$)))

(defun coerce-string (&rest parms)
  (eval 
  `(concatenate 'string ,@(loop for x in parms collect (to-string x)))))
   
(defun show-raam (raam &key (refresh t))
  (when refresh
	(forget-trees raam)
	(loop for x in (raam-tset raam) do (encode raam x)))
  (let* ((symtab (append (raam-symtab0 raam) (raam-symtab1 raam))))
    (format t "Raam~%")
    (format t "  Termtest = ~a~%" (raam-termtest raam))
    (format t "  Symtab~%")
    (show-trees 
     (loop for x in symtab collect (car x))
     (loop for x in symtab collect (cadr x)))))

(defun show-trees (trees vecs)
  (let* ((n (max-length trees))
	 (f$ (coerce-string "     ~" n "a = ~a~%")))
    (loop for tree in trees
	  for vec in vecs
	  do
	  (format t f$ tree (format-vector vec)))))
    


;;; ---------------------------------------------------------------------------
;;;
;;; Evaluation
;;;
;;; ---------------------------------------------------------------------------

(defun same-structure? (t1 t2)
  (or (and (atom t1) (atom t2))
      (and (listp t1)
	   (listp t2)
	   (= (length t1) (length t2))
	   (same-structure? (car t1) (car t2))
	   (same-structure? (cdr t1) (cdr t2)))))

(defun visual-compare (r1 r2 &aux (sametreecount 0) (samestructcount 0) (n (length r1)))
  (unless (= n (length r2)) (error "lists must be same size to compare"))
  (loop for x1 in r1 
	for x2 in r2
	for sametree?   = (equalp x1 x2)
	for samestruct? = (same-structure? x1 x2) do
	(format t " ~3a,~3a: ~a~%      --> ~a~%~%"
		sametree? samestruct? x1 x2)
	(if sametree?   (incf sametreecount))
	(if samestruct? (incf samestructcount)))
  (format t "~a trees, ~a completely correct, ~a structurally correct"
	  (length r1) sametreecount samestructcount)
  (= sametreecount (length r1)))

(defun representable? (raam tree-list &key
			    (max-depth (max-depth tree-list))
			    (print t)
			    (debug nil))
  (if print 
      (let* ((res (loop for tree in tree-list collect
			(rdecode raam (encode raam tree)
				 :debug debug
				 :max-depth max-depth))))
	(visual-compare tree-list res))
    (loop for tree in tree-list
	  always 
	  (equal tree (rdecode raam (encode raam tree)
			       :debug debug
			       :max-depth max-depth)))))

(defun reconstruct-trees (raam &key max-depth (print t))
  (let* ((trees (raam-tset raam)))
    (unless max-depth (setf max-depth (max-depth trees)))
    (representable? raam trees 
		    :max-depth max-depth
		    :print print)))
