;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;;
;;;     			     ANSIL
;;; 		      Advanced Network Simulator In Lisp
;;; 		      ----------------------------------
;;;
;;;				  Written By
;;;			       Peter J. Angeline
;;;			      Gregory M. Saunders
;;;
;;;			   The Ohio State University
;;;		       Laboratory for AI Research (LAIR)
;;;			     Columbus, Ohio, 43210
;;;
;;;			   Copyright (c) 1991, 1992
;;;
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; ---------------------------------------------------------------------- 
;;; File: show.lisp - Collection of functions to show ansil data structures.
;;;
;;; Author: Gregory M Saunders
;;;
;;; License: This code may be distributed free of charge.  This code may not be
;;; incorporated into any production software of any type or any code which is
;;; sold without the express written consent of the authors.  Modifications and
;;; inclusions for other types of software are permitted, provided that this
;;; header remains unchanged and included in all subsequent versions of the
;;; code.
;;;
;;; Disclaimer: The authors of this code make no claims as to its
;;; validity, and are in no way responsible for any errors contained within.
;;; 
;;; File description: This file contains an implementation of back prop using
;;; linear algebra routines rather than symbolic means.  
;;;
;;; Creation date: March 11, 1992, Version 2.00
;;; Modifications: none 
;;; ----------------------------------------------------------------------
;;;
;;;
;;;                         Change Log
;;;                         ----------
;;;
;;; ----------------------------------------------------------------------


;;; ----------------------------------------------------------------------
;;;
;;; Matrices
;;;
;;; ----------------------------------------------------------------------

(defun show-char (ch n)
  (loop for i from 1 to n do (format t "~a" ch))
  (format t "~%"))

(defun abstract (x threshold max min)
  (cond ((< x min) '-)
	((> x max) '+)
	((< (- x min) threshold) min)
	((< (- max x) threshold) max)
	(t x)))

;; &allow-other-keys is included so that other routines can call show-matrix
;; and pass paramters easily.  For example, see "show-raam"
;; gms 01/21
(defun show-matrix (m &key (digits 3) (width (+ digits 4)) (threshold zero)
		      (max 100) (min -100) (free-weights 'all) &allow-other-keys)
  "Prints a matrix, rounding entries to max/min if they're within threshold"
  (let* ((d (array-dimensions m))
	 (r (length d)))
    (cond ((= r 1) (format t "  ~a~%" m))
	  ((= r 2)
	   (let* ((max-row (1- (car d)))
		  (max-col (1- (cadr d)))
		  (width$  (princ-to-string width))
		  (wid$    (princ-to-string (1+ width)))
		  (digits$ (princ-to-string digits))
		  (count$  (concatenate 'string "~" wid$ "@a"))
		  (array$  (concatenate 'string "~" width$ "," digits$ "f")))
	     (format t "~%  f\\t|")
	     (loop for i from 0 to max-col do (format t count$ i))
	     (format t "~%  ---\\-")
	     (show-char "-" (* (1+ max-col) (1+ width)))
	     (loop for i from 0 to max-row do
		   (format t "  ~2a | " i)
		   (loop for j from 0 to max-col 
			 for num1 = (aref m i j)
			 for num2 = (abstract num1 threshold max min)
			 for fixed = (if (equal free-weights 'all) nil
				       (= (aref free-weights i j) zero))
			 for ch = (if (and fixed
					   (not (zerop num1)))
				      "*" " ")
			 do
			 (format t array$ 
				 (if (or (not fixed)
					 (not (= num1 zero)))
				     num2 ""))
			 (format t "~a" ch))
		   (terpri))
	     (format t "~%")))
	  (t (error "Can only print matrices of dimension 2 or less")))))



;;; ---------------------------------------------------------------------------
;;;
;;; Networks
;;;
;;; ---------------------------------------------------------------------------

(defmacro show-network (net &rest parms &key (verbose nil)
			    &allow-other-keys
			    &aux (f$ "  ~11a = ~a~%"))
  `(progn
     (when ,verbose
	   (format t "~a" (set-network ,net)))
     (format t "Nodes:~%")
     (format t ,f$ "Activations" (format-vector (network-activations ,net)))
     (format t ,f$ "Deltas"      (format-vector (network-deltas ,net)))
     (format t "~%~%Weights (\"*\" = fixed weight)~%")
     (show-network-weights ,net ,@parms)))

(defmacro show-network-weights (net &rest parms)
  `(show-matrix (network-weights ,net)
		:free-weights (network-free-weights ,net)
		,@parms))

(defun show-patterns (net &key (patterns (network-tset net)) verbose)
  (loop for p in patterns
	for x    = (car p)
	for out1 = (fpass net x)
	for out2 = (if verbose out1 (network-outputs net))
	for targ = (cadr p)
	do
	(format t "  Input  = ~a~%  Target = ~a~%  Output = ~a~%~%"
		(format-vector x)
		(format-vector targ)
		(if verbose (format-vector out2) out2))))


;;; ---------------------------------------------------------------------------
;;;
;;; Generic show
;;;
;;; ---------------------------------------------------------------------------

(defun show (x &rest parms &key &allow-other-keys &aux (args (cons x parms)))
  (typecase x 
	    (vector  (apply #'show-vector  args))
	    (matrix  (apply #'show-matrix  args))
	    (network (eval `(show-network ,@args)))
	    (raam    (apply #'show-raam    args))
	    (list    (loop for y in x do (apply #'show (cons y parms))))
	    (t       (apply #'show-user args))))

(defun show-user (x &rest parms &key &allow-other-keys &aux (args (cons x parms)))
  (declare (ignore args))
  (typecase x
	    (t (format t "~a~%" x))))
